#include "perlInter.h"


/* #define _DEBUG_  */

#ifdef _DEBUG_
#define D(x) printf("%s\n",x);
#else
#define D(x)
#endif

static PerlInterpreter *interp = NULL;
static int    		refcnt=0;
char   			errbuf[256];

char *perl_error()
{
	return errbuf;
}


/* create a perl interpreter */
void perl_create()
{
	char *embedding[] = {"", "-e", "0" };

	if (refcnt == 0) {
		D("creating perl interpreter")
		interp = perl_alloc();
		perl_construct( interp );
		perl_parse( interp, NULL, 3, embedding, NULL );
		perl_run( interp );
		}
	refcnt++;
}


/* destroy a perl interpreter */
void perl_destroy()
{
	if (refcnt == 1) {
		D("destroying perl interpreter")
		perl_destruct( interp );
		perl_free( interp );
		}
	refcnt--;
}

/* evaluates a perl script */
void perl_eval(char *script)
{
  	SV* sv =  newSVpv(script, strlen(script));
	perl_eval_sv( sv, FALSE );
	sv_clear( sv );
}





/* returns the value of a perl variable */
/*
Since perl uses scalars you need not worry about datatypes
when calling the Int,Float and String methods. But with the
List and dict methods the name of the perl script variable must be
a List or Dict data type respectively.
*/

/* helper functions */

/* forward declarations */
PyObject *cvt(SV *sv);
PyObject *cvt_list(AV *av);
PyObject *cvt_dict(HV *hv);


int safe_get(char *name, SV **sv)
{
	if ((*sv = perl_get_sv( name, FALSE )) == NULL) {
		sprintf(errbuf,"Error in perlInter: Unable to resolve scalar '%s'",
			name );
		return -1;
		}
	return 0;		
}

PyObject *cvt_dict(HV *tb)
{
	PyObject *result, *p;
	I32  max, i, retlen;
	HE *he;

	max  = hv_iterinit( tb );
	result = PyDict_New();

	for (i = 0; i < max; i++) {
		char *ckey;
		SV *value;	
		PyObject *key;
	
		/* get the next key value tuple */
		he =  hv_iternext( tb );	
		if (he == NULL) {
			sprintf(errbuf,"hv_iternext returned NULL in cvt_dict");
			return Py_None;
			}
		
		ckey  = hv_iterkey(he, &retlen);
		if (ckey == NULL) {
			sprintf(errbuf,"hv_interkey returned NULL in cv_dict");
			return Py_None;
			}

		key   = PyString_FromString( ckey );
		value =  hv_iterval( tb, he );
		if (value == NULL) {
			sprintf(errbuf,"hv_interval returned NULL in cv_dict");
			return Py_None;
			}

		if ((p = cvt(value)) != Py_None) {
			D("before")
			PyDict_SetItem(result, key, p );
			D("after")
			}		
		}
	return result;
}

PyObject *cvt_list(AV *av)
{
	PyObject *result, *p;
	int max, i;
	
	max = av_len(av);
	result = PyList_New( 0 );
	for (i = 0; i <= max; i++) {
		SV** sv = av_fetch(av, (I32) i, 0);
		if (*sv == NULL) {
			sprintf(errbuf,"Inside cvt_list, av_fetch(av,%d) returned NULL",
				i);
			return Py_None;
			}	
		if ((p = cvt(*sv)) != Py_None)	
			PyList_Append(result, p);
		}
	return result;
}



/* converts a perl SV into a PyObject */
PyObject *cvt(SV *sv)
{
	PyObject *result;
	int type;
	
	type = SvTYPE( sv );

	switch ( type )
	{
	/* atomic types */ 
	case SVt_IV:
		return PyInt_FromLong( (long) SvIV( sv ) );
	case SVt_NV: 
		D("converting a realnumber")
		return PyFloat_FromDouble( (double) SvNV( sv ) );
	case SVt_PV: 
	case SVt_PVIV:
	case SVt_PVNV:
		{
		char *p;
		if ((p = SvPV( sv, na )) == NULL) {
			sprintf(errbuf,"SvPV failed in cvt");
			return Py_None;
			}
		return PyString_FromString( p );
		}
	/* data structures */
	case SVt_PVAV: 
		D("converting a list")
		return cvt_list( (AV *) sv );
	case SVt_PVHV: 
		D("converting a dictionary")
		return cvt_dict( (HV *) sv );
	default:
		sprintf(errbuf,"Unsupported type '%d'",type);
		break; 
	}
	return Py_None;
}



int perl_IntFromName(char *name)
{
	SV *sv;
	if (safe_get(name, &sv)) return 0;
	return SvIV(sv);
}

float perl_FloatFromName(char *name)
{
	SV *sv;
	if (safe_get(name, &sv)) return 0;
	return SvNV( sv );
}

char *perl_StringFromName(char *name)
{
	SV *sv;
	if (safe_get(name, &sv)) return 0;
	return SvPV( sv, na );
}


PyObject *perl_ListFromName(char *name)
{
	 AV  *av = perl_get_av( name , FALSE);
	 
	 /* error checking */
	 if (av == NULL) {
	 	sprintf(errbuf,"Error in perlInter: Unable to resolve array named '%s'",
	 		name );
	 	return Py_None;
	 	}
	return cvt_list( av );
}

PyObject *perl_DictFromName(char *name)
{
	HV *hv = perl_get_hv( name, FALSE );
	 /* error checking */
	 if (hv == NULL) {
	 	sprintf(errbuf,"Error in perlInter: Unable to resolve dictionary named '%s'",
	 		name );
	 	return Py_None;
	 	}
	return cvt_dict( hv );
}

